/***************************  BOSE.C  *****************************
 * Title: Bose-Nelson Sort Pair Generator
 * Author: Joe Celko
 * Acknowledgment: adapted from Rich, Robert P., INTERNAL SORTING
 *    METHODS ILLUSTRATED  WITH PL/I PROGRAMS, Prentice-Hall, 1973.
 * Compiler: TC++ 1.0, Microsoft C 6.0
 * Compile time switches:
 *      SHOWSTACK: if defined, stack is displayed at intervals.
 *
 * Usage: bose n outfile
 *
 * Generates swap pairs for a sort of size n, writing them to outfile.
 * (c) 1990 C Gazette. Use freely but acknowledge author/publication
 *******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* common variables */
int count = 0;          /* running total of number of swap pairs */
FILE *OutFile;          /* write swap pairs to this file */

/* we do a lot of work on a stack */
#define MAXSTACK 2500   /* pick a large number for stack size */
int top = 0;            /* stack index */
int stack[MAXSTACK];    /* stack */

/* prototypes */

void BoseSort ( int n );
float OptimalSwap ( int n ); /* optional information */
void Push5 ( int a, int b, int c, int d, int e);
void Push3 ( int a, int b, int c );
int Even (int e );
#ifdef SHOWSTACK
void ShowStack ( void );
#endif

main ( int argc, char *argv[] )
{
    int n;
    char Usage[]="USAGE: bose n outfile, where n > 0\n";

    if ( argc < 3 ) 
    {
        puts( Usage );
        exit ( 0 );
    }
    else 
    {
        n = atoi ( argv[1] );  /* get element count */
        if ( n <= 0 ) 
        {
            puts ( Usage );
            exit ( 1 );
        }
        if (( OutFile = fopen ( argv[2], "w" )) == NULL ) 
        {
            printf ( "Can't open %s\n", argv[2] );
            exit ( 1 );
        }
        BoseSort ( n );
        fprintf( OutFile, "} /* There were %d swaps */\n", count );
        printf ( " There were %d swaps generated.\n", count );
        printf ( " The optimal sort has %.0f swaps.\n", OptimalSwap(n));
    }
}

void BoseSort(int n)
{
    int a, b;       /* middle points in intervals */
    int i, x, j, y; /* these are defined for easy reading */

    fprintf( OutFile,
             "{ /* Bose-Nelson sort for %d elements */\n", n );
    Push5 ( 2, 1, n, 0, 0 );  /* load stack with starting values */
    while ( top > 0 ) 
    {
        switch ( stack[top] ) 
        {
          case 0: /* end of swap list */
            top = 0;
            break;

          case 1:
            {
                fprintf ( OutFile, 
                  "swap(%d, %d);\n", stack[top-1], stack[top-2]) ;
                top = top - 3;
                count++;
            } 
            break;

          case 2:
            {
                i = stack[top - 1];
                x = stack[top - 2];
                top = top - 3;
                if ( x != 1 ) 
                {
                    a = ( x/2 );
                    Push5 ( 3, i, a, ( i+a ), ( x-a ));
                    Push3 ( 2, ( i+a ), ( x-a ));
                    Push3 ( 2, i, a );
                }
            } 
            break;

          case 3:
            {
                i = stack[top - 1];
                x = stack[top - 2];
                j = stack[top - 3];
                y = stack[top - 4];
                a = ( x/2 );
                b = ( Even(x)) ? ( (y+1)/2 ) : ( y/2 );
                top = top - 5;

                if ( (x == 1) && (y == 1) )
                    Push3 ( 1, i, j );
                else if ( (x == 1) && (y == 2) ) 
                {
                    Push3 ( 1, i, j );
                    Push3 ( 1, i, (j+1) );
                }
                else if ((x == 2) && (y == 1)) 
                {
                    Push3 ( 1, (i+1), j );
                    Push3 ( 1, i, j );
                }
                else 
                {
                    Push5 ( 3, ( i+a ), (x-a ), j, b);
                    Push5 ( 3, ( i+a ), ( x-a ), ( j+b ), ( y-b ));
                    Push5 ( 3, i, a, j, b );
                }
            } 
            break;

          default:
            {
                puts ( "FATAL: error in stack\n" );
                exit ( 2 );
            } 
            break;
        } /* end of switch */
#ifdef SHOWSTACK
        ShowStack(); /* Optional: Display stack in operation */
#endif
    } /* end of while loop */
}     /* end of main */

int Even ( int e )
{
    return( 1 - ( 1 & e ));
}

void Push3 ( int a, int b, int c )
{
    if ( top + 3 >= MAXSTACK ) 
    {
        puts ( "FATAL: stack overflow\n" );
        exit ( 3 );
    };
    stack[++top] = c;
    stack[++top] = b;
    stack[++top] = a;
}

void Push5 ( int a, int b, int c, int d, int e )
{
    if ( top+5 >= MAXSTACK ) 
    {
        puts ( "FATAL: stack overflow\n" );
        exit ( 5 );
    };
    stack[++top] = e;
    stack[++top] = d;
    stack[++top] = c;
    stack[++top] = b;
    stack[++top] = a;
}

#ifdef SHOWSTACK
void ShowStack(void)
/* optional procedure to display stack during running */
{
    int p;
    p = top;
    while ( p > 0 )
        switch ( stack[p] ) 
        {
          case 0: /* end of swap list */
            {
                puts( "\n\n hit any key to continue." );
                p = getch();
                p = 0;
            } 
            break;

          case 1:
            {
                printf ( "P(%d, %d, %d); swap pair\n ", 
                         stack[p], stack[p-1], stack[p-2] );
                p = p - 3;
            } 
            break;
          case 2:
            {
                printf ( "P(%d, %d, %d);\n", 
                         stack[p], stack[p-1], stack[p-2] );
                p = p - 3;
            } 
            break;

          case 3:
            {
                printf ( "P(%d, %d, %d, %d, %d);\n",
                    stack[p], stack[p-1], stack[p-2], 
                    stack[p-3], stack[p-4] );
                p = p - 5;
            } 
            break;

          default:
            {
                puts( "FATAL: error in stack\n" );
                while ( p > 0 )
                    printf ( "stack[%d] = %d;\n", p, stack[p] );
                exit( 1 );
            } 
            break;
        } /* end of switch */
}
#endif

float OptimalSwap ( int n )
{
    /* Computes the ceil(log2(n!)).
       Uses these formula:
              (n!) = n * (n-1) * ... * 1
              logA(x*y) = logA(x) + logA(y);
              logA(x) = logB(x) / logB(A)
              log10(2) = .30103;
    */
    int i;
    float result = 0.0;
    for ( i = 1; ( i <= n ); i++ ) 
        result = result + ( log10 ( i ) / 0.30103 );

    return ( ceil ( result ));
}